unit snappy;

interface

type
  TSnappyStatus = (
    SNAPPY_OK = 0,
    SNAPPY_INVALID_INPUT = 1,
    SNAPPY_BUFFER_TOO_SMALL = 2
  );

var

// * Takes the data stored in "input[0..input_length-1]" and stores
// * it in the array pointed to by "compressed".
// *
// * <compressed_length> signals the space available in "compressed".
// * If it is not at least equal to "snappy_max_compressed_length(input_length)",
// * SNAPPY_BUFFER_TOO_SMALL is returned. After successful compression,
// * <compressed_length> contains the true length of the compressed output,
// * and SNAPPY_OK is returned.
// *
// * Example:
// *   size_t output_length = snappy_max_compressed_length(input_length);
// *   char* output = (char*)malloc(output_length);
// *   if (snappy_compress(input, input_length, output, &output_length)
// *       == SNAPPY_OK) {
// *     ... Process(output, output_length) ...
// *   }
// *   free(output);
  snappy_compress: function(const input; input_length: Cardinal;
    var compressed; var compressed_length: Cardinal): TSnappyStatus; cdecl;


// * Given data in "compressed[0..compressed_length-1]" generated by
// * calling the snappy_compress routine, this routine stores
// * the uncompressed data to
// *   uncompressed[0..uncompressed_length-1].
// * Returns failure (a value not equal to SNAPPY_OK) if the message
// * is corrupted and could not be decrypted.
// *
// * <uncompressed_length> signals the space available in "uncompressed".
// * If it is not at least equal to the value returned by
// * snappy_uncompressed_length for this stream, SNAPPY_BUFFER_TOO_SMALL
// * is returned. After successful decompression, <uncompressed_length>
// * contains the true length of the decompressed output.
// *
// * Example:
// *   size_t output_length;
// *   if (snappy_uncompressed_length(input, input_length, &output_length)
// *       != SNAPPY_OK) {
// *     ... fail ...
// *   }
// *   char* output = (char*)malloc(output_length);
// *   if (snappy_uncompress(input, input_length, output, &output_length)
// *       == SNAPPY_OK) {
// *     ... Process(output, output_length) ...
// *   }
// *   free(output);
  snappy_uncompress: function(const compressed; compressed_length: Cardinal;
    var uncompressed; var uncompressed_length: Cardinal): TSnappyStatus; cdecl;


// * Returns the maximal size of the compressed representation of
// * input data that is "source_length" bytes in length.
  snappy_max_compressed_length: function(source_length: Cardinal): Cardinal; cdecl;


// * REQUIRES: "compressed[]" was produced by snappy_compress()
// * Returns SNAPPY_OK and stores the length of the uncompressed data in
// * *result normally. Returns SNAPPY_INVALID_INPUT on parsing error.
// * This operation takes O(1) time.
  snappy_uncompressed_length: function(const compressed; compressed_length: Cardinal;
                                      var resultLength: Cardinal): TSnappyStatus; cdecl;


// * Check if the contents of "compressed[]" can be uncompressed successfully.
// * Does not return the uncompressed data; if so, returns SNAPPY_OK,
// * or if not, returns SNAPPY_INVALID_INPUT.
// * Takes time proportional to compressed_length, but is usually at least a
// * factor of four faster than actual decompression.
  snappy_validate_compressed_buffer: function(const compressed;
                                        compressed_length: Cardinal): TSnappyStatus; cdecl;


  function LoadLibrary: Boolean;
  procedure UnloadLibrary;

implementation

uses SysUtils, Windows;

var
  _HLib: THandle;

function GetProcAddress(const procName: PChar): Pointer;
begin
  Result := Windows.GetProcAddress(_HLib, procName);
  if Result = nil then
    raise Exception.CreateFmt('Failed to get procedure %s address', [String(procName)]);
end;

function LoadLibrary: Boolean;
const
  LibName = 'snappy.dll';
begin
  if _HLib <> 0 then
  begin
    Result := true; // library is already loaded
    Exit;
  end;
  _HLib := Windows.LoadLibrary(LibName);
  if _HLib = 0 then raise Exception.CreateFmt('Could not load library: %s', [LibName]);

  snappy_compress := GetProcAddress('snappy_compress');
  snappy_uncompress := GetProcAddress('snappy_uncompress');
  snappy_max_compressed_length := GetProcAddress('snappy_max_compressed_length');
  snappy_uncompressed_length := GetProcAddress('snappy_uncompressed_length');
  snappy_validate_compressed_buffer := GetProcAddress('snappy_validate_compressed_buffer');

  Result := True;
end;

procedure UnloadLibrary;
begin
  Windows.FreeLibrary(_HLib);
end;

end.
